#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import NoReturn, Union, Tuple
from pydantic.dataclasses import dataclass
import numpy
import matplotlib.pyplot as plt
import shapely.geometry as geo
from matplotlib.patches import PathPatch
from matplotlib import patches
from matplotlib.path import Path
from matplotlib.collections import PatchCollection
from LightWave2D.grid import Grid
from shapely.affinity import scale, rotate


# Configuration dictionary for dataclasses
config_dict = {
    'kw_only': True,
    'extra': 'forbid',
    'slots': True,
    'arbitrary_types_allowed': True
}


@dataclass(kw_only=True, config=config_dict)
class BaseComponent():
    """
    Represents a generic component in a simulation grid.

    Attributes:
        grid (Grid): The simulation grid.
        facecolor (str): The color of the component's face.
        edgecolor (str): The color of the component's edge.
        alpha (float): Transparency level of the component.
        rotation (float): Rotation angle of the component.
        sigma (float): Conductivity of the component.
    """
    grid: Grid
    facecolor: str = 'lightblue'
    edgecolor: str = 'blue'
    alpha: float = 0.3
    rotation: float = 0
    sigma: float = 0

    def __post_init__(self):
        x0, y0 = self.position
        self.coordinate = self.grid.get_coordinate(x=x0, y=y0)

        self.build_object()

    def build_object(self) -> NoReturn:
        """
        Build the permittivity mesh for the component.
        """
        self.epsilon_r_mesh = numpy.ones(self.grid.shape)  # Background permittivity

        self.compute_polygon()

        self.polygon = rotate(self.polygon, angle=self.rotation)

        self.polygon = self.polygon.intersection(self.grid.polygon)

        self.path = Path(self.polygon.exterior.coords)

        coordinates = numpy.c_[self.grid.x_mesh.T.flatten(), self.grid.y_mesh.T.flatten()]

        self.idx = self.path.contains_points(coordinates).astype(bool).reshape(self.epsilon_r_mesh.shape)

    def add_to_ax(self, ax: plt.axis) -> PatchCollection:
        """
        Add the component to the provided axis.

        Args:
            ax (plt.Axes): The axis to which the component will be added.

        Returns:
            PatchCollection: The collection of patches added to the axis.
        """
        path = Path.make_compound_path(
            Path(numpy.asarray(self.polygon.exterior.coords)[:, :2]),
            *[Path(numpy.asarray(ring.coords)[:, :2]) for ring in self.polygon.interiors])

        patch = PathPatch(path, facecolor=self.facecolor, edgecolor=self.edgecolor, alpha=0.4)
        collection = PatchCollection([patch], facecolor=self.facecolor, edgecolor=self.edgecolor, alpha=self.alpha)

        ax.add_collection(collection, autolim=True)
        ax.autoscale_view()
        return collection

    def plot(self) -> NoReturn:
        """
        Plot the component.
        """
        figure, ax = plt.subplots(1, 1, figsize=(6, 6))
        ax.set_title('FDTD Simulation')
        ax.set_xlabel(r'x position [$\mu$m]')
        ax.set_xlabel(r'y position [$\mu$m]')
        ax.set_aspect('equal')

        self.add_to_ax(ax)
        ax.autoscale_view()

        plt.show()

    def add_to_epsilon_r_mesh(self, epsilon_r_mesh: numpy.ndarray) -> NoReturn:
        """
        Add the component's permittivity to the provided mesh.

        Args:
            epsilon_r_mesh (numpy.ndarray): The permittivity mesh to be updated.
        """
        epsilon_r_mesh[self.idx] = self.epsilon_r

    def add_to_sigma_mesh(self, sigma_mesh: numpy.ndarray) -> NoReturn:
        """
        Add the component's sigma to the provided mesh.

        Args:
            sigma_mesh (numpy.ndarray): The permittivity mesh to be updated.
        """
        sigma_mesh[self.idx] = self.sigma

    def add_non_linear_effect_to_field(self, field: numpy.ndarray) -> NoReturn:
        """
        Add non-linear effects to the field.

        Args:
            field (numpy.ndarray): The field to which non-linear effects will be added.
        """
        return field
        chi_3 = 1e-8  # Third-order non-linear susceptibility (example value)
        intensity = numpy.abs(field) ** 2
        nonlinear_term = chi_3 * intensity * field
        return field + self.grid.dt * nonlinear_term


@dataclass(config=config_dict)
class Waveguide(BaseComponent):
    """
    Represents a Waveguide scatterer in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): The center position of the scatterer.
        epsilon_r (float): The relative permittivity inside the scatterer.
        side_length (float): The side length of the square scatterer.
    """
    position_0: Tuple[Union[float, str], Union[float, str]]
    position_1: Tuple[Union[float, str], Union[float, str]]
    width: float
    epsilon_r: float

    def __post_init__(self):
        x0, y0 = self.position_0
        x1, y1 = self.position_1
        self.coordinate_0 = self.grid.get_coordinate(x=x0, y=y0)
        self.coordinate_1 = self.grid.get_coordinate(x=x1, y=y1)

        self.build_object()

    def compute_polygon(self) -> Path:
        """
        Compute the polygon of the square scatterer.
        """
        line = geo.LineString(
            [(self.coordinate_0.x, self.coordinate_0.y), (self.coordinate_1.x, self.coordinate_1.y)]
        )

        self.polygon = line.buffer(self.width)


@dataclass(config=config_dict)
class Square(BaseComponent):
    """
    Represents a square scatterer in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): The center position of the scatterer.
        epsilon_r (float): The relative permittivity inside the scatterer.
        side_length (float): The side length of the square scatterer.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    side_length: float

    def compute_polygon(self) -> Path:
        """
        Compute the polygon of the square scatterer.
        """
        half_side = self.side_length / 2
        x_start, x_end = self.coordinate.x - half_side, self.coordinate.x + half_side
        y_start, y_end = self.coordinate.y - half_side, self.coordinate.y + half_side

        self.polygon = geo.Polygon([
            (x_start, y_start),
            (x_start, y_end),
            (x_end, y_end),
            (x_end, y_start)
        ]).convex_hull


@dataclass(config=config_dict)
class Circle(BaseComponent):
    """
    Represents a circular scatterer in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): The center position of the scatterer.
        epsilon_r (float): The relative permittivity inside the scatterer.
        radius (float): The radius of the scatterer.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    radius: float

    def compute_polygon(self) -> NoReturn:
        """
        Compute the polygon of the circular scatterer.
        """
        self.polygon = geo.Point(self.coordinate.x, self.coordinate.y).buffer(self.radius)


@dataclass(config=config_dict)
class Triangle(BaseComponent):
    """
    Represents a triangular scatterer in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): The center position of the scatterer.
        epsilon_r (float): The relative permittivity inside the scatterer.
        side_length (float): The side length of the triangular scatterer.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    side_length: float

    def compute_polygon(self) -> NoReturn:
        """
        Compute the polygon of the triangular scatterer.
        """
        height = (numpy.sqrt(3) / 2) * self.side_length

        # Vertices of an equilateral triangle centered at (coordinate.x, coordinate.y)
        p0 = (self.coordinate.x - self.side_length / 2, self.coordinate.y - height / 3)
        p1 = (self.coordinate.x + self.side_length / 2, self.coordinate.y - height / 3)
        p2 = (self.coordinate.x, self.coordinate.y + 2 * height / 3)

        self.polygon = geo.Polygon([p0, p1, p2])


@dataclass(config=config_dict)
class Ellipse(BaseComponent):
    """
    Represents an elliptical scatterer in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): The center position of the scatterer.
        width (float): The width of the ellipse.
        height (float): The height of the ellipse.
        epsilon_r (float): The relative permittivity inside the scatterer.
        resolution (float): The resolution of the ellipse.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    width: float
    height: float
    epsilon_r: float
    resolution: float = 100

    def compute_polygon(self) -> NoReturn:
        """
        Compute the polygon of the elliptical scatterer.
        """
        ellipse_patch = patches.Ellipse(
            (self.coordinate.x, self.coordinate.y),
            width=self.resolution,
            height=self.resolution * self.height / self.width,
        )

        vertices = ellipse_patch.get_verts()
        polygon = geo.Polygon(vertices)

        self.polygon = scale(polygon, xfact=self.width / self.resolution, yfact=self.width / self.resolution)


@dataclass(config=config_dict)
class Grating(BaseComponent):
    """
    Represents a grating in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): The starting position of the grating.
        epsilon_r (float): The relative permittivity inside the grating.
        period (float): The period of the grating.
        duty_cycle (float): The duty cycle of the grating.
        num_periods (int): The number of periods in the grating.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    period: float
    duty_cycle: float
    num_periods: int

    def compute_polygon(self) -> NoReturn:
        """
        Compute the polygon of the grating.
        """
        bars = []
        for i in range(self.num_periods):
            x_start = self.coordinate.x + i * self.period
            x_end = x_start + self.duty_cycle * self.period
            bar = geo.box(x_start, self.coordinate.y - 0.5, x_end, self.coordinate.y + 0.5)
            bars.append(bar)

        self.polygon = geo.MultiPolygon(bars)


@dataclass(config=config_dict)
class RingResonator(BaseComponent):
    """
    Represents a ring resonator in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): The center position of the ring resonator.
        epsilon_r (float): The relative permittivity inside the resonator.
        inner_radius (float): The inner radius of the ring.
        width (float): The width of the ring.
    """
    grid: Grid
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    inner_radius: float
    width: float

    def compute_polygon(self) -> NoReturn:
        """
        Compute the polygon of the ring resonator.
        """
        inner_circle = geo.Point(self.coordinate.x, self.coordinate.y).buffer(self.inner_radius)
        outer_circle = geo.Point(self.coordinate.x, self.coordinate.y).buffer(self.inner_radius + self.width)
        self.polygon = outer_circle.difference(inner_circle)


@dataclass(config=config_dict)
class Lense(BaseComponent):
    """
    Represents a lens in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): The center position of the lens.
        epsilon_r (float): The relative permittivity inside the lens.
        width (float): The width of the lens.
        curvature (float): The curvature of the lens (positive for convex, negative for concave).
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    width: float
    curvature: float

    def compute_polygon(self) -> NoReturn:
        """
        Compute the polygon of the lens.
        """
        # Create two arcs for the lens shape
        x0 = self.coordinate.x + self.curvature - self.width / 2
        x1 = self.coordinate.x - self.curvature + self.width / 2

        arc3 = geo.Point(x0, self.coordinate.y).buffer(self.curvature, resolution=100)
        arc4 = geo.Point(x1, self.coordinate.y).buffer(self.curvature, resolution=100)

        self.polygon = arc3.intersection(arc4)

# -



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import NoReturn
import numpy as np
import numpy
from LightWave2D.grid import Grid
from pydantic.dataclasses import dataclass
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt

# Configuration dictionary for dataclasses
config_dict = {
    'kw_only': True,
    'slots': True,
    'extra': 'forbid',
    'arbitrary_types_allowed': True
}


@dataclass(config=config_dict)
class PML:
    """
    Represents a Perfectly Matched Layer (PML) for absorbing boundary conditions in FDTD simulations.

    Attributes:
        grid (Grid): The simulation grid.
        width (int): Width of the PML region in grid cells.
        sigma_max (float): Maximum value of the conductivity profile.
        order (int): Polynomial order of the conductivity profile.
    """
    grid: Grid
    width: str = 10
    sigma_max: float = 0.045
    order: int = 3

    def __post_init__(self):
        """
        Initialize the PML regions with appropriate conductivity profiles.
        """
        self.sigma_x = np.zeros((self.grid.n_x, self.grid.n_y))
        self.sigma_y = np.zeros((self.grid.n_x, self.grid.n_y))

        y_mesh, x_mesh = numpy.meshgrid(self.grid.y_stamp, self.grid.x_stamp)

        opposite_width = float(self.width.strip("%"))
        opposite_width = f"{100 - opposite_width}%"
        self.width_start = self.grid.get_coordinate(x=self.width, y=self.width)
        self.width_stop = self.grid.get_coordinate(x=opposite_width, y=opposite_width)

        bottom_boundary = y_mesh < self.width_start.y
        left_boundary = x_mesh < self.width_start.x

        top_boundary = y_mesh > self.width_stop.y
        right_boundary = x_mesh > self.width_stop.x

        self.sigma_y[bottom_boundary] = self.sigma_max * ((self.width_start.y - y_mesh[bottom_boundary]) / self.width_start.y) ** self.order
        self.sigma_y[top_boundary] = self.sigma_max * ((y_mesh[top_boundary] - self.width_stop.y) / self.width_start.y) ** self.order
        self.sigma_x[left_boundary] = self.sigma_max * ((self.width_start.x - x_mesh[left_boundary]) / self.width_start.x) ** self.order
        self.sigma_x[right_boundary] = self.sigma_max * ((x_mesh[right_boundary] - self.width_stop.x) / self.width_start.x) ** self.order

    def add_to_ax(self, ax: plt.Axes) -> NoReturn:
        """
        Add the PML regions to a matplotlib axis.

        Args:
            ax (plt.Axes): The axis to which the PML regions will be added.
        """
        cmap = np.zeros([256, 4])
        cmap[:, 3] = np.linspace(0, 1, 256)
        cmap = ListedColormap(cmap)

        ax.pcolormesh(
            self.grid.x_stamp,
            self.grid.y_stamp,
            self.sigma_y.T + self.sigma_x.T,
            cmap=cmap,
        )

    def plot(self, unit_size: int = 6) -> NoReturn:
        """
        Plot the PML regions.

        Args:
            unit_size (int): Size of each unit in the plot.
        """
        figsize = int(unit_size), int(unit_size * self.grid.size_y / self.grid.size_x)
        figure, ax = plt.subplots(1, 1, figsize=figsize)

        ax.set_title('FDTD Simulation PML Regions')
        ax.set_ylabel(r'y position [$\mu$m]')
        ax.set_xlabel(r'x position [$\mu$m]')
        ax.set_aspect('equal')

        self.add_to_ax(ax)
        plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy
from typing import Tuple, NoReturn, List, Union
from LightWave2D.utils import bresenham_line
from pydantic.dataclasses import dataclass
from matplotlib.path import Path
import shapely.geometry as geo
import matplotlib.pyplot as plt
from LightWave2D.physics import Physics
from matplotlib.patches import PathPatch
import matplotlib
from matplotlib.collections import PatchCollection
from LightWave2D.grid import Grid


# Configuration dictionary for dataclasses
config_dict = {
    'kw_only': True,
    'extra': 'forbid',
    'slots': True,
    'arbitrary_types_allowed': True
}


@dataclass(kw_only=True, config=config_dict)
class BaseSource:
    """
    Base class for sources in a 2D light wave simulation.

    Attributes:
        grid (Grid): The grid of the simulation mesh.
        facecolor (str): Color of the source face.
        edgecolor (str): Color of the source edge.
        alpha (float): Transparency level of the source.
        rotation (float): Rotation angle of the source.
    """
    grid: Grid
    facecolor: str = 'red'
    edgecolor: str = 'red'
    alpha: float = 0.3
    rotation: float = 0

    def __post_init__(self):
        self.coordinate = self.grid.get_coordinate(x=self.position[0], y=self.position[1])
        self.build_object()

    def build_object(self) -> NoReturn:
        """
        Build the source object and initialize its properties.
        """
        self.compute_polygon()
        self.path = Path(self.polygon.exterior.coords)
        self.path = self.path.transformed(matplotlib.transforms.Affine2D().rotate_around(self.coordinate.x, self.coordinate.y, self.rotation))
        x_mesh, y_mesh = numpy.meshgrid(self.grid.x_stamp, self.grid.y_stamp)
        coordinates = numpy.c_[x_mesh.T.flatten(), y_mesh.T.flatten()]
        self.idx = self.path.contains_points(coordinates).astype(bool).reshape(self.grid.shape)

    def add_to_ax(self, ax: plt.Axes) -> PatchCollection:
        """
        Add the source to the provided axis as a scatter point or patch.

        Args:
            ax (plt.Axes): The axis to which the source will be added.
        """
        path = Path.make_compound_path(
            Path(numpy.asarray(self.polygon.exterior.coords)[:, :2]),
            *[Path(numpy.asarray(ring.coords)[:, :2]) for ring in self.polygon.interiors]
        )
        patch = PathPatch(path, facecolor=self.facecolor, edgecolor=self.edgecolor, alpha=self.alpha)
        collection = PatchCollection([patch], facecolor=self.facecolor, edgecolor=self.edgecolor, alpha=self.alpha)
        ax.add_collection(collection, autolim=True)
        ax.autoscale_view()
        return collection

    def plot(self) -> NoReturn:
        """
        Plot the source on a 2D grid.
        """
        figure, ax = plt.subplots(1, 1, figsize=(6, 6))
        ax.set_title('FDTD Simulation')
        ax.set_xlabel(r'x position [$\mu$m]')
        ax.set_ylabel(r'y position [$\mu$m]')
        ax.set_aspect('equal')
        self.add_to_ax(ax)
        plt.show()


@dataclass(config=config_dict)
class PointSource(BaseSource):
    """
    Represents a point source in a 2D light wave simulation.

    Attributes:
        wavelength (Union[float, List[float], numpy.ndarray]): Wavelength of the source.
        position (Tuple[Union[float, str], Union[float, str]]): Position (x, y) of the source.
        amplitude (float): Amplitude of the electric field.
    """
    wavelength: Union[float | List[float] | numpy.ndarray]
    position: Tuple[float | str, float | str]
    amplitude: float

    def __post_init__(self):
        self.wavelength = numpy.atleast_1d(self.wavelength)
        self.frequency = Physics.c / self.wavelength
        self.omega = 2 * numpy.pi * self.frequency
        x, y = self.position
        self.p0 = self.grid.get_coordinate(x=x, y=y)
        self.polygon = geo.Point(self.p0.x, self.p0.y)
        self.path = Path(self.polygon.coords)

    def add_to_ax(self, ax: plt.Axes) -> NoReturn:
        """
        Add the point source to the provided axis as a scatter point.

        Args:
            ax (plt.Axes): The axis to which the point source will be added.
        """
        ax.scatter(
            self.p0.x,
            self.p0.y,
            color=self.facecolor,
            label='source'
        )

    def add_source_to_field(self, field: numpy.ndarray, time: float) -> NoReturn:
        """
        Add the source's effect to the simulation field.

        Args:
            field (numpy.ndarray): The simulation field to which the source's effect will be added.
            time (float): The current simulation time.
        """
        field[self.p0.x_index, self.p0.y_index] = self.amplitude / len(self.omega) * numpy.sin(self.omega * time).sum()


@dataclass(config=config_dict)
class PointImpulsion(BaseSource):
    """
    Represents a point impulsion source in a 2D light wave simulation.

    Attributes:
        amplitude (float): Amplitude of the electric field.
        duration (float): Duration of the impulsion.
        position (Tuple[Union[float, str], Union[float, str]]): Position (x, y) of the source.
        delay (float): Delay before the impulsion starts.
    """
    amplitude: float
    duration: float
    position: Tuple[float | str, float | str]
    delay: float = 0

    def __post_init__(self):
        x, y = self.position
        self.p0 = self.grid.get_coordinate(x=x, y=y)
        self.polygon = geo.Point(self.p0.x, self.p0.y)
        self.path = Path(self.polygon.coords)

    def add_to_ax(self, ax: plt.Axes) -> NoReturn:
        """
        Add the point source to the provided axis as a scatter point.

        Args:
            ax (plt.Axes): The axis to which the point source will be added.
        """
        ax.scatter(
            self.p0.x,
            self.p0.y,
            color=self.facecolor,
            label='source'
        )

    def add_source_to_field(self, field: numpy.ndarray, time: float) -> NoReturn:
        """
        Add the impulsion source's effect to the simulation field.

        Args:
            field (np.ndarray): The simulation field to which the source's effect will be added.
            time (float): The current simulation time.
        """
        source_field = numpy.exp(-((time - self.delay) / self.duration) ** 2)

        field[self.p0.x_index, self.p0.y_index] = self.amplitude * source_field


@dataclass(config=config_dict)
class LineSource(BaseSource):
    """
    Represents a line source in a 2D light wave simulation.

    Attributes:
        amplitude (float): Amplitude of the electric field.
        wavelength (float): Wavelength of the source.
        position_0 (Tuple[Union[float, str], Union[float, str]]): Starting position (x, y) of the source.
        position_1 (Tuple[Union[float, str], Union[float, str]]): Ending position (x, y) of the source.
    """
    amplitude: float
    wavelength: float
    position_0: Tuple[float | str, float | str]
    position_1: Tuple[float | str, float | str]

    def __post_init__(self):
        self.frequency = Physics.c / self.wavelength
        self.omega = 2 * numpy.pi * self.frequency
        self.build_object()

    def build_object(self) -> NoReturn:
        """
        Build the line source object and calculate the line's coordinates.
        """
        self.p0 = self.grid.get_coordinate(x=self.position_0[0], y=self.position_0[1])
        self.p1 = self.grid.get_coordinate(x=self.position_1[0], y=self.position_1[1])

        position = bresenham_line(
            x0=self.p0.x_index,
            y0=self.p0.y_index,
            x1=self.p1.x_index,
            y1=self.p1.y_index
        )

        rows, cols = zip(*position.T)
        self.slice_indexes = rows, cols

        p0 = geo.Point(self.p0.x, self.p0.y)
        p1 = geo.Point(self.p1.x, self.p1.y)
        self.polygon = geo.LineString((p0, p1))

    def add_to_ax(self, ax: plt.axis) -> NoReturn:
        """
        Add the scatterer to the provided axis as a circle.

        Args:
            ax (Axis): The axis to which the scatterer will be added.
        """
        ax.plot(
            self.polygon.xy[0],
            self.polygon.xy[1],
            color=self.facecolor,
            label='source'
        )

    def add_source_to_field(self, field: numpy.ndarray, time: float) -> NoReturn:
        """
        Add the line source's effect to the simulation field.

        Args:
            field (np.ndarray): The simulation field to which the source's effect will be added.
            time (float): The current simulation time.
        """
        rows, cols = self.slice_indexes
        field[rows, cols] = self.amplitude * numpy.sin(self.omega * time)


@dataclass(config=config_dict)
class LineImpulsion(BaseSource):
    """
    Represents a line impulsion source in a 2D light wave simulation.

    Attributes:
        duration (float): Duration of the impulsion.
        amplitude (float): Amplitude of the electric field.
        position_0 (Tuple[Union[float, str], Union[float, str]]): Starting position (x, y) of the source.
        position_1 (Tuple[Union[float, str], Union[float, str]]): Ending position (x, y) of the source.
        delay (float): Delay before the impulsion starts.
    """
    duration: float
    amplitude: float
    position_0: Tuple[float | str, float | str]
    position_1: Tuple[float | str, float | str]
    delay: float = 0

    def __post_init__(self):
        self.build_object()

    def build_object(self) -> NoReturn:
        """
        Build the line source object and calculate the line's coordinates.
        """
        self.p0 = self.grid.get_coordinate(x=self.position_0[0], y=self.position_0[1])
        self.p1 = self.grid.get_coordinate(x=self.position_1[0], y=self.position_1[1])

        position = bresenham_line(
            x0=self.p0.x_index,
            y0=self.p0.y_index,
            x1=self.p1.x_index,
            y1=self.p1.y_index
        )

        rows, cols = zip(*position.T)
        self.slice_indexes = rows, cols

        p0 = geo.Point(self.p0.x, self.p0.y)
        p1 = geo.Point(self.p1.x, self.p1.y)
        self.polygon = geo.LineString((p0, p1))

    def add_to_ax(self, ax: plt.axis) -> NoReturn:
        """
        Add the scatterer to the provided axis as a circle.

        Args:
            ax (Axis): The axis to which the scatterer will be added.
        """
        ax.plot(
            self.polygon.xy[0],
            self.polygon.xy[1],
            color=self.facecolor,
            label='source'
        )

    def add_source_to_field(self, field: numpy.ndarray, time: float) -> NoReturn:
        """
        Add the line impulsion source's effect to the simulation field.

        Args:
            field (np.ndarray): The simulation field to which the source's effect will be added.
            time (float): The current simulation time.
        """
        rows, cols = self.slice_indexes

        source_field = numpy.exp(-((time - self.delay) / self.duration) ** 2)

        field[rows, cols] = self.amplitude * source_field


# -



"""
Experiment: Circular Scatterer
==============================

This example demonstrates the setup and execution of a circular scatterer experiment using LightWave2D.
We will define the simulation grid, add a circular scatterer and a line source, apply a perfectly matched layer (PML), run the simulation, and visualize the results.
"""

# %%
# Importing the necessary packages
from LightWave2D.grid import Grid
from LightWave2D.experiment import Experiment
from MPSPlots import colormaps

# %%
# Define the simulation grid
grid = Grid(
    resolution=0.1e-6,  # Grid resolution in meters
    size_x=32e-6,       # Grid size in the x direction in meters
    size_y=20e-6,       # Grid size in the y direction in meters
    n_steps=300         # Number of time steps for the simulation
)

# Initialize the experiment with the defined grid
experiment = Experiment(grid=grid)

# %%
# Add a circular scatterer to the experiment
scatterer = experiment.add_circle(
    position=('30%', '50%'),  # Center position of the scatterer
    epsilon_r=1.5,            # Relative permittivity of the scatterer
    radius=3e-6,              # Radius of the circular scatterer in meters
    # sigma=1e6
)

# %%
# Add a line source to the experiment
source = experiment.add_line_source(
    wavelength=1550e-9,         # Wavelength of the source in meters
    position_0=('10%', '80%'),  # Starting position of the source
    position_1=('10%', '20%'),  # Ending position of the source
    amplitude=10                # Amplitude of the source
)

# %%
# Add a perfectly matched layer (PML) to absorb boundary reflections
experiment.add_pml(
    order=1,          # Order of the PML polynomial profile
    width='10%',      # Width of the PML region as a percentage of grid size
    sigma_max=5000    # Maximum conductivity for the PML
)

# %%
# Plot the entire experiment setup
experiment.plot()

# Run the FDTD simulation
experiment.run_fdtd()

# %%
# Plot the last time frame of the computed fields
experiment.plot_frame(
    frame_number=-1,  # Plot the last frame
    scale_max=1,      # Maximum scale for the field visualization
    colormap=colormaps.polytechnique.red_black_blue,  # Colormap for the plot
    # show_intensity=True
)

# %%
# Render an animation of the field propagation over time
animation = experiment.render_propagation(
    skip_frame=5,                            # Number of frames to skip in the animation
    unit_size=5,                             # Size of each unit in the animation
    colormap=colormaps.polytechnique.red_black_blue,  # Colormap for the animation
    scale_max=1,
    auto_adjust_clim=True
)

# Save the animation as a GIF file
animation.save('./circular_scatterer_propagation.gif', writer='Pillow', fps=10)
